//****************************************************************************************
//                                 R K 2 _ L A T _ L O N                                 *
//****************************************************************************************

const double mu = 398600.4418;  // Earth's gravitational parameter (km^3/s^2)
double latitude, longitude;
struct State {
    double x, y, z;
    double vx, vy, vz;
};
State acceleration(const State& state) {
    double r = std::sqrt(state.x * state.x + state.y * state.y + state.z * state.z);
    double factor = -mu / (r * r * r);
    State acc;
    acc.x = factor * state.x;
    acc.y = factor * state.y;
    acc.z = factor * state.z;
    return acc;
}
void rungeKutta2ndOrder(State& state, double dt) {
    State k1 = acceleration(state);
    State halfStepState;
    halfStepState.x = state.x + 0.5 * state.vx * dt;
    halfStepState.y = state.y + 0.5 * state.vy * dt;
    halfStepState.z = state.z + 0.5 * state.vz * dt;
    halfStepState.vx = state.vx + 0.5 * k1.x * dt;
    halfStepState.vy = state.vy + 0.5 * k1.y * dt;
    halfStepState.vz = state.vz + 0.5 * k1.z * dt;
    State k2 = acceleration(halfStepState);
    state.x += state.vx * dt + 0.5 * k1.x * dt * dt;
    state.y += state.vy * dt + 0.5 * k1.y * dt * dt;
    state.z += state.vz * dt + 0.5 * k1.z * dt * dt;
    state.vx += 0.5 * (k1.x + k2.x) * dt;
    state.vy += 0.5 * (k1.y + k2.y) * dt;
    state.vz += 0.5 * (k1.z + k2.z) * dt;
}
State satellite;
void xyzToLatLon(double x, double y, double z, double& latitude, double& longitude) {
    double r = std::sqrt(x * x + y * y + z * z);
    latitude = std::asin(z / r) * RAD_TO_DEG;
    longitude = std::atan2(y, x) * RAD_TO_DEG;
};

//****************************************************************************************
//                                     G L O B A L S                                     *
//****************************************************************************************
#include "assets.h"
#include <ArduinoJson.h>
StaticJsonDocument<10000> satdoc;    /* allocated memory of 10000 was found to be enough but less may work */
StaticJsonDocument<10000> planetdoc; /* allocated memory of 10000 was found to be enough but less may work */
int longitude_pixel, latitude_pixel, EXTRA_API_DELAY;
const long INTERVAL_ANIM = 50;
const long INTERVAL_API = 1000*30;
unsigned long previousMillisAPI, previousMillisAnim;
int XHAIR_COLOR = 0xbf7d; // https://rgbcolorpicker.com/565
int animRings = 3; 


//****************************************************************************************
//                                        W I F I                                        *
//****************************************************************************************
#include <WiFiManager.h> /* this must be included first becuase of a bug with TFT_eSPI + WiFiManager */
#include <WiFi.h>
#include <HTTPClient.h>
String getRequest(String url)
{
  HTTPClient http;http.begin(url);
  int httpCode = http.GET();if(httpCode == HTTP_CODE_OK) {
    String payload = http.getString();
    return payload;}  
}

//****************************************************************************************
//                                         T F T                                         *
//****************************************************************************************
#include <TFT_eSPI.h>
TFT_eSPI tft             = TFT_eSPI();
TFT_eSprite mainSpr      = TFT_eSprite(&tft);
TFT_eSprite xHairSrp     = TFT_eSprite(&tft); // lat lon cross hair
TFT_eSprite comspr       = TFT_eSprite(&tft); // lat lon cross hair

#define Main_SPRITE_WIDTH      320
#define Main_SPRITE_HEIGHT     170
#define BACKLIGHT_PIN     8
#define CF_OL32 &Orbitron_Light_32
#define GFXFF 1

struct Pixel {
    int x;
    int y;
};
void latLonToPixel(Pixel& pixel, double newLatitude, double newLongitude); // i have no idea why https://stackoverflow.com/questions/62683653/variable-or-field-declared-void-error-on-ardunio-compiler
void latLonToPixel(Pixel& pixel, double newLatitude, double newLongitude) {
    int newLonInt = static_cast<int>((worldMap_WIDTH / 2) + newLongitude * (worldMap_WIDTH_F / 2) / 180.0);
    int newLatInt = static_cast<int>((worldMap_HEIGHT / 2) - newLatitude * (worldMap_HEIGHT_F / 2) / 90.0);
    pixel.x = newLonInt;
    pixel.y = newLatInt;
}
Pixel llPixel;

void setXHairGraphic(const char* name){
  xHairSrp.fillSprite(TFT_TRANSPARENT);
  xHairSrp.fillCircle(0, 0, 3, XHAIR_COLOR);
  xHairSrp.drawCircle(0, 0, 5, XHAIR_COLOR);
  xHairSrp.drawLine(0, 0-10, 0, 0+10, XHAIR_COLOR);
  xHairSrp.drawLine(0-10, 0, 0+10, 0, XHAIR_COLOR);  
  comspr.setTextColor(XHAIR_COLOR, TFT_TRANSPARENT); 
  comspr.fillSprite(TFT_TRANSPARENT);
  comspr.drawString(name, comspr.width() / 2, 10, GFXFF); 
}
const unsigned long planetInterval = 15000; // 15 seconds in milliseconds
unsigned long planetPreviousMillis = 0;
int currentPlanet = 0; //
#define ISS_BUTTON 16
#define PLANETS_BUTTON 5
int currentFunction = 1;
unsigned long lastDebounceTime = 0;
const unsigned long debounceDelay = 500; // 500ms debounce time
void checkButtonPress(const unsigned long cMillis) {
    if (cMillis - lastDebounceTime > debounceDelay) {
        if (digitalRead(ISS_BUTTON) == LOW) {
          currentFunction = 1;
          xyzToLatLon(satellite.x,satellite.y,satellite.z, latitude, longitude);
          latLonToPixel(llPixel,latitude,longitude);
          XHAIR_COLOR = 0xbf7d;
          setXHairGraphic("ISS");
          setNeoPixelToHex565(XHAIR_COLOR);
          lastDebounceTime = cMillis;
        } 
        else if (digitalRead(PLANETS_BUTTON) == LOW) {
          lastDebounceTime = cMillis;
          if (currentFunction == 2){
            // currentPlanet = (currentPlanet + 1) % 11;
            planetPreviousMillis = - planetInterval;
            return;
          }
          currentFunction = 2;
          currentPlanet = 10;
          planetPreviousMillis = cMillis;
          changePlanet("sun",0xff80);
        } 
        // else if (digitalRead(button3) == LOW) {
        //     currentFunction = 3;
        //     Serial.println("Switched to Function 3");
        //     lastDebounceTime = millis();
        // }
    }
}

//****************************************************************************************
//                                   N E O _ P I X E L                                   *
//****************************************************************************************
#include <Adafruit_NeoPixel.h>
#define LED_PIN    37
#define LED_COUNT 1
Adafruit_NeoPixel strip(LED_COUNT, LED_PIN, NEO_GRB + NEO_KHZ400);
int neo_r;
int neo_g;
int neo_b;
void setNeoPixelToHex565(int hex){
  neo_r = ((hex >> 11) & 0x1F) * 255 / 31;
  neo_g = ((hex >> 5) & 0x3F) * 255 / 63;
  neo_b = (hex & 0x1F) * 255 / 31;
  strip.setPixelColor(0, neo_r/30, neo_g/30, neo_b/30);
  // Serial.print(neo_r/30);Serial.print(" ");Serial.print(neo_g/30);Serial.print(" ");Serial.println(neo_b/30);
  strip.show();
}

//****************************************************************************************
//                                      S E T _ U P                                      *
//****************************************************************************************

void changePlanet(const char* planetName, int newColor)
{
  const double lat  = planetdoc[planetName]["deg_lat"];
  const double lon  = planetdoc[planetName]["deg_lon"];
  const double dist = planetdoc[planetName]["km_distance"];
  // const char* body  = planetdoc[planetName]["body"];
  XHAIR_COLOR = newColor;
  latLonToPixel(llPixel,lat,lon);
  setXHairGraphic(planetName);
  setNeoPixelToHex565(XHAIR_COLOR);
}

void setup() {

  Serial.begin(115200);
  // while (!Serial) {;}
  Serial.println("Serial initialized.");

  strip.begin();
  strip.show();

  tft.begin();
  tft.setRotation(1);
  tft.fillScreen(TFT_BLACK);
  pinMode(BACKLIGHT_PIN, OUTPUT);
  digitalWrite(BACKLIGHT_PIN, HIGH);
  mainSpr.createSprite(Main_SPRITE_WIDTH,Main_SPRITE_HEIGHT);
  mainSpr.fillScreen(TFT_TRANSPARENT);
  mainSpr.setSwapBytes(true);
  mainSpr.pushImage(0,0,worldMap_WIDTH,worldMap_HEIGHT,worldMap);
  mainSpr.pushSprite(160-worldMap_WIDTH/2,85-worldMap_HEIGHT/2,TFT_TRANSPARENT);
  xHairSrp.createSprite(20, 20); // xHairSrp.setColorDepth(8);  //8 bit color-depth to save PSRAM
  xHairSrp.setOrigin(10, 10);
  xHairSrp.fillSprite(TFT_TRANSPARENT);
  comspr.setFreeFont(CF_OL32);  
  comspr.setTextDatum(MC_DATUM);
  comspr.createSprite(160, 40);

  
  WiFiManager wm;
  // wm.resetSettings(); use this to reset chip for a new router
  bool res;
  res = wm.autoConnect("PlanetLocationAP","password");
  if(!res) {Serial.println("Failed to connect");/* ESP.restart(); */} 
  else {Serial.println("connected...yeey :)");}
  delay(500);


  String payLoad = getRequest("https://orbital-api.com/satelliteVectorEciOnEcef?norad=25544&basic=true");
  DeserializationError error = deserializeJson(satdoc, payLoad);    
  satellite.x = satdoc["sv"][0]; 
  satellite.y = satdoc["sv"][1]; 
  satellite.z = satdoc["sv"][2]; 
  satellite.vx = satdoc["sv"][3];
  satellite.vy = satdoc["sv"][4];
  satellite.vz = satdoc["sv"][5];

  currentFunction = 1;
  xyzToLatLon(satellite.x,satellite.y,satellite.z, latitude, longitude);
  latLonToPixel(llPixel,latitude,longitude);
  XHAIR_COLOR = 0xbf7d;
  setXHairGraphic("ISS");
  setNeoPixelToHex565(XHAIR_COLOR);

  String payLoad2 = getRequest("https://orbital-api.com/allPlanetsLatLonDist");
  DeserializationError error2 = deserializeJson(planetdoc, payLoad2);   

  previousMillisAPI = millis();
  EXTRA_API_DELAY = 0;

  pinMode(ISS_BUTTON, INPUT); // TX QTPY ESP32-S2
  pinMode(PLANETS_BUTTON, INPUT); // RX QTPY ESP32-S2
}

//****************************************************************************************
//                                        L O O P                                        *
//****************************************************************************************


void planetChangeCheck(unsigned long cMillis){
  if (cMillis - planetPreviousMillis >= planetInterval) {
    planetPreviousMillis = cMillis; // Reset timer
    currentPlanet = (currentPlanet + 1) % 11; // Cycle through 11 functions
    // Serial.print("Switched to Planet ");
    // Serial.println(currentPlanet);
    switch (currentPlanet) {
        case 0:  changePlanet("mercury",0xe77d); break;
        case 1:  changePlanet("venus",0xee92); break;
        case 2:  changePlanet("earth",0x1685); break;
        case 3:  changePlanet("moon",TFT_LIGHTGREY); break;
        case 4:  changePlanet("mars",0xe282); break;
        case 5:  changePlanet("jupiter",0xdca3); break;
        case 6:  changePlanet("saturn",0xe631); break;
        case 7:  changePlanet("uranus",0xbf7d); break;
        case 8:  changePlanet("neptune",0x3add); break;
        case 9:  changePlanet("pluto",0xe75a); break;
        case 10: changePlanet("sun",0xff80); break;
    }
  }
}

double deltaTime; 

void loop() {
  unsigned long currentMillis = millis();
  checkButtonPress(currentMillis);
  if (currentMillis - previousMillisAnim >= INTERVAL_ANIM) {

    deltaTime = (currentMillis - previousMillisAnim) / 1000.0;
    previousMillisAnim = currentMillis;
    //physics
    
    rungeKutta2ndOrder(satellite, deltaTime);
    // xyzToLatLon(satellite.x,satellite.y,satellite.z, latitude, longitude);
    // latLonToPixel(llPixel,latitude,longitude);
    switch (currentFunction) {
        case 1:
          // rungeKutta2ndOrder(satellite, 0.05);
          xyzToLatLon(satellite.x,satellite.y,satellite.z, latitude, longitude);
          latLonToPixel(llPixel,latitude,longitude);
          break;
        case 2:
          planetChangeCheck(currentMillis);
          break;
    }
    //graphics
    mainSpr.fillSprite(TFT_BLACK);
    mainSpr.setSwapBytes(true);
    mainSpr.pushImage(0,0,worldMap_WIDTH,worldMap_HEIGHT,worldMap);
    mainSpr.setSwapBytes(false);
    comspr.pushToSprite(&mainSpr,Main_SPRITE_WIDTH/2-80/2,Main_SPRITE_HEIGHT-comspr.height()-4,TFT_TRANSPARENT);
    xHairSrp.pushToSprite(&mainSpr,
      (-10)+llPixel.x,
      (-10)+llPixel.y,
      TFT_TRANSPARENT); 
    animRings += 1;
    animRings = animRings%10;
    for(int i=0;i<4;i+=1){
      mainSpr.drawCircle(
        llPixel.x, 
        llPixel.y, 
        animRings+i*10, 
        XHAIR_COLOR);}
    mainSpr.pushSprite(160-worldMap_WIDTH/2,85-worldMap_HEIGHT/2);
  }

  if (currentMillis - previousMillisAPI >= INTERVAL_API) {
    previousMillisAPI = currentMillis;
    EXTRA_API_DELAY += 1;
    if (EXTRA_API_DELAY == 20){
      EXTRA_API_DELAY = 0;
      String payLoad = getRequest("https://orbital-api.com/satelliteVectorEciOnEcef?norad=25544&basic=true");
      DeserializationError error = deserializeJson(satdoc, payLoad);    
      satellite.x = satdoc["sv"][0]; 
      satellite.y = satdoc["sv"][1]; 
      satellite.z = satdoc["sv"][2]; 
      satellite.vx = satdoc["sv"][3];
      satellite.vy = satdoc["sv"][4];
      satellite.vz = satdoc["sv"][5];
      String payLoad2 = getRequest("https://orbital-api.com/allPlanetsLatLonDist");
      DeserializationError error2 = deserializeJson(planetdoc, payLoad2);  
      Serial.println("fresh data!!!");
    }
  }
}